"""Unit tests for reviewboard.accounts.models.Profile."""

from __future__ import annotations

from typing import TYPE_CHECKING

from django_assert_queries.testing import assert_queries
from django.core.cache import cache
from django.contrib.auth.models import AnonymousUser, User
from django.db.models import Count, Q, Value
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.models import LocalSiteProfile, Profile
from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from django_assert_queries import ExpectedQueries


class ProfileTests(TestCase):
    """Unit tests for reviewboard.accounts.models.Profile."""

    fixtures = ['test_users']

    def test_get_display_name_unauthenticated_public(self):
        """Testing Profile.get_display_name with a public profile"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = False

        self.assertEqual(profile.get_display_name(AnonymousUser()),
                         user.username)

    def test_get_display_name_unauthenticated_private(self):
        """Testing Profile.get_display_name for an unauthenticated user viewing
        a private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(profile.get_display_name(AnonymousUser()),
                         user.username)

    def test_get_display_name_public(self):
        """Testing Profile.get_display_name for an authenticated user viewing a
        public profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = False

        self.assertEqual(
            profile.get_display_name(User.objects.get(username='grumpy')),
            user.get_full_name())

    def test_get_display_name_private(self):
        """Testing Profile.get_display_name for an authenticated user viewing a
        private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(
            profile.get_display_name(User.objects.get(username='grumpy')),
            user.username)

    def test_get_display_name_admin_private(self):
        """Testing Profile.get_display_name for an admin viewing a private
        profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(
            profile.get_display_name(User.objects.get(username='admin')),
            user.get_full_name())

    @add_fixtures(['test_site'])
    def test_get_display_name_localsite_member_private(self):
        """Testing Profile.get_display_name for a LocalSite member viewing
        a LocalSite member with a private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        viewer = User.objects.get(username='grumpy')
        site = LocalSite.objects.get(pk=1)
        site.users.add(viewer)

        self.assertEqual(profile.get_display_name(viewer), user.username)

    @add_fixtures(['test_site'])
    def test_get_display_name_localsite_admin_private(self):
        """Testing Profile.get_display_name for a LocalSite admin viewing
        a LocalSite member with a private profile
        """
        user = User.objects.get(username='admin')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(
            profile.get_display_name(User.objects.get(username='doc')),
            user.get_full_name())

    @add_fixtures(['test_site'])
    def test_get_display_name_localsite_admin_private_other_site(self):
        """Testing Profile.get_display_name for a LocalSite admin viewing a
        member of another LocalSite with a private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        viewer = User.objects.get(username='grumpy')
        site = LocalSite.objects.create(name='site-3')
        site.users.add(viewer)
        site.admins.add(viewer)

        self.assertEqual(profile.get_display_name(viewer), user.username)

    def test_get_display_name_self_private(self):
        """Testing Profile.get_display_name for a user viewing themselves with
        a private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(profile.get_display_name(user), user.get_full_name())

    def test_star_review_group(self):
        """Testing Profile.star_review_group"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        group1 = self.create_review_group(name='group1')
        profile.starred_groups.add(group1)

        group2 = self.create_review_group(name='group2')

        # 1 query:
        #
        # 2. Insert-or-ignore new entry
        queries = [
            {
                'type': 'INSERT',
                'model': Profile.starred_groups.through,
            },
        ]

        with assert_queries(queries):
            profile.star_review_group(group2)

        self.assertEqual(list(profile.starred_groups.order_by('pk')),
                         [group1, group2])

    @add_fixtures(['test_site'])
    def test_star_review_group_invalidates_cache(self):
        """Testing Profile.star_review_group invalidates count caches"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)
        local_site.users.add(user)

        profile.star_review_group(self.create_review_group(name='group1'))

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review group count.
        cross_site_queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                1)

        # Fetch the count for the global site.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review group count.
        global_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'annotations': {'__count': Count('*')},
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 1)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 1)

        # Fetch the count for the LocalSite.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review group count.
        local_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'annotations': {'__count': Count('*')},
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': (Q(starred_by__id=user.pk) &
                          Q(local_site=local_site)),
            },
        ]

        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                0)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                0)

        # Star again. This should invalidate the global and cross-site caches.
        profile.star_review_group(self.create_review_group(name='group2'))
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                2)

        # Fetch the count for the global site.
        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 2)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 2)

        # Fetch the count for the LocalSite. This should still be in cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                0)

    @add_fixtures(['test_site'])
    def test_star_review_group_with_local_site_invalidates_cache(self):
        """Testing Profile.star_review_group with LocalSite invalidates
        count caches
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)
        local_site.users.add(user)

        profile.star_review_group(self.create_review_group(
            name='group1',
            local_site=local_site))

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review group count.
        cross_site_queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                1)

        # Fetch the count for the LocalSite.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review group count.
        local_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'annotations': {'__count': Count('*')},
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                1)

        # Fetch the count for the global site.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review group count.
        global_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'annotations': {'__count': Count('*')},
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 0)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 0)

        # Star again. This should invalidate the LocalSite and cross-site
        # caches.
        profile.star_review_group(self.create_review_group(
            name='group2',
            local_site=local_site))
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                2)

        # Fetch the count for the LocalSite.
        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                2)

        # Fetch the count for the global site. This should still be in cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(),
                0)

    def test_unstar_review_group(self):
        """Testing Profile.unstar_review_group"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        group = self.create_review_group(name='group1')
        profile.starred_groups.add(group)

        # 2 queries:
        #
        # 1. Fetch existing starred entries (to avoid duplicates)
        # 2. Remove the entry
        queries = [
            {
                'model': Profile.starred_groups.through,
                'where': Q(profile=(profile.pk,)) & Q(group__in={group.pk}),
            },
            {
                'type': 'DELETE',
                'model': Profile.starred_groups.through,
                'where': Q(id__in=[group.pk]),
            },
        ]

        with assert_queries(queries):
            profile.unstar_review_group(group)

        self.assertFalse(profile.starred_groups.exists())
        self.assertEqual(
            profile.star_cache,
            {
                (Group, 1): False,
            })

    @add_fixtures(['test_site'])
    def test_unstar_review_group_invalidates_cache(self):
        """Testing Profile.unstar_review_group invalidates count caches"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)
        local_site.users.add(user)

        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2')
        profile.star_review_group(group1)
        profile.star_review_group(group2)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        #
        # 1 query:
        #
        # 1. Fetch user's starred group count.
        cross_site_queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                2)

        # Fetch the count for the global site.
        #
        # 1 query:
        #
        # 1. Fetch user's starred group count.
        global_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 2)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 2)

        # Fetch the count for the LocalSite.
        #
        # 1 query:
        #
        # 1. Fetch user's starred group count.
        local_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                0)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                0)

        # Unstar the group. This should invalidate the global and cross-site
        # caches.
        profile.unstar_review_group(group2)

        self.assertTrue(LocalSite.objects.has_local_sites())
        self.assertEqual(
            profile.star_cache,
            {
                (Group, 1): True,
                (Group, 2): False,
            })

        # Fetch the count for cross-sites.
        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                1)

        # Fetch the count for the global site.
        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 1)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 1)

        # Fetch the count for the LocalSite. This should still be in cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                0)

    @add_fixtures(['test_site'])
    def test_unstar_review_group_with_local_site_invalidates_cache(self):
        """Testing Profile.unstar_review_group with LocalSite invalidates
        count caches
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)
        local_site.users.add(user)

        group1 = self.create_review_group(name='group1',
                                          local_site=local_site)
        group2 = self.create_review_group(name='group2',
                                          local_site=local_site)
        profile.star_review_group(group1)
        profile.star_review_group(group2)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        #
        # 1 query:
        #
        # 1. Fetch user's starred group count.
        cross_site_queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                2)

        # Fetch the count for the LocalSite.
        #
        # 1 query:
        #
        # 1. Fetch user's starred group count.
        local_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                2)

        # Fetch the count for the global site.
        #
        # 1 query:
        #
        # 1. Fetch user's starred group count.
        global_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 0)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 0)

        # Unstar the group. This should invalidate the LocalSite and
        # cross-site caches.
        profile.unstar_review_group(group2)

        self.assertTrue(LocalSite.objects.has_local_sites())
        self.assertEqual(
            profile.star_cache,
            {
                (Group, 1): True,
                (Group, 2): False,
            })

        # Fetch the count for cross-sites.
        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                1)

        # Fetch the count for the LocalSite.
        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                1)

        # Fetch the count for the global site. This should still be in cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 0)

    def test_star_review_request_with_discarded(self):
        """Testing Profile.star_review_request with discarded review request"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(local_site=None)

        review_request1 = self.create_review_request()
        profile.starred_review_requests.add(review_request1)

        review_request2 = self.create_review_request(publish=True)
        review_request2.close(ReviewRequest.DISCARDED)

        # 1 query:
        #
        # 1. Insert-or-ignore new entry
        queries = [
            {
                'type': 'INSERT',
                'model': Profile.starred_review_requests.through,
            },
        ]

        with assert_queries(queries):
            profile.star_review_request(review_request2)

        self.assertEqual(list(profile.starred_review_requests.order_by('pk')),
                         [review_request1, review_request2])
        self.assertEqual(site_profile.starred_public_request_count, 0)

    def test_star_review_request_with_draft(self):
        """Testing Profile.star_review_request with draft review request"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(local_site=None)

        review_request1 = self.create_review_request()
        profile.starred_review_requests.add(review_request1)

        review_request2 = self.create_review_request()

        # 1 query:
        #
        # 1. Insert-or-ignore new entry
        queries = [
            {
                'type': 'INSERT',
                'model': Profile.starred_review_requests.through,
            },
        ]

        with assert_queries(queries):
            profile.star_review_request(review_request2)

        self.assertEqual(list(profile.starred_review_requests.order_by('pk')),
                         [review_request1, review_request2])
        self.assertEqual(site_profile.starred_public_request_count, 0)

    def test_star_review_request_with_pending(self):
        """Testing Profile.star_review_request with pending review request"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(local_site=None)

        review_request1 = self.create_review_request()
        profile.starred_review_requests.add(review_request1)
        site_profile.starred_public_request_count = 1
        site_profile.save(update_fields=('starred_public_request_count',))

        review_request2 = self.create_review_request(publish=True)

        # 3 queries:
        #
        # 1. Insert-or-ignore new entry
        # 2. Update LocalSiteProfile.starred_public_request_count
        # 3. Re-fetch LocalSiteProfile.starred_public_request_count
        queries = [
            {
                'type': 'INSERT',
                'model': Profile.starred_review_requests.through,
            },
            {
                'type': 'UPDATE',
                'model': LocalSiteProfile,
                'where': Q(pk=user.pk),
            },
            {
                'model': LocalSiteProfile,
                'values_select': ('starred_public_request_count',),
                'where': Q(pk=user.pk),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            profile.star_review_request(review_request2)

        self.assertEqual(list(profile.starred_review_requests.order_by('pk')),
                         [review_request1, review_request2])
        self.assertEqual(site_profile.starred_public_request_count, 2)

    def test_star_review_request_with_submitted(self):
        """Testing Profile.star_review_request with submitted review request"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(local_site=None)

        review_request1 = self.create_review_request()
        profile.starred_review_requests.add(review_request1)
        site_profile.starred_public_request_count = 1
        site_profile.save(update_fields=('starred_public_request_count',))

        review_request2 = self.create_review_request(publish=True)
        review_request2.close(ReviewRequest.SUBMITTED)

        # 3 queries:
        #
        # 1. Insert-or-ignore new entry
        # 2. Update LocalSiteProfile.starred_public_request_count
        # 3. Re-fetch LocalSiteProfile.starred_public_request_count
        queries = [
            {
                'type': 'INSERT',
                'model': Profile.starred_review_requests.through,
            },
            {
                'type': 'UPDATE',
                'model': LocalSiteProfile,
                'where': Q(pk=user.pk),
            },
            {
                'model': LocalSiteProfile,
                'values_select': ('starred_public_request_count',),
                'where': Q(pk=user.pk),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            profile.star_review_request(review_request2)

        self.assertEqual(list(profile.starred_review_requests.order_by('pk')),
                         [review_request1, review_request2])
        self.assertEqual(site_profile.starred_public_request_count, 2)

    @add_fixtures(['test_site'])
    def test_star_review_request_invalidates_cache(self):
        """Testing Profile.star_review_request invalidates count cache"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)
        local_site.users.add(user)

        profile.star_review_request(self.create_review_request())

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        cross_site_queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                1)

        # Fetch the count for the global site.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        global_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 1)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 1)

        # Fetch the count for the LocalSite.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        local_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                0)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                0)

        # Star again. This should invalidate the global and cross-site caches.
        profile.star_review_request(self.create_review_request(publish=True))

        # Fetch the count for cross-sites.
        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                2)

        # Fetch the count for the global site.
        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 2)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 2)

        # Fetch the count for the LocalSite. This should still be in cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                0)

    @add_fixtures(['test_site'])
    def test_star_review_request_with_local_site_invalidates_cache(self):
        """Testing Profile.star_review_request with LocalSite invalidates
        count cache
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)
        local_site.users.add(user)

        profile.star_review_request(
            self.create_review_request(local_site=local_site))

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        cross_site_queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                1)

        # Fetch the count for the LocalSite.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        local_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                1)

        # Fetch the count for the global site.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        global_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

        # Star again. This should invalidate the global and cross-site caches.
        profile.star_review_request(
            self.create_review_request(publish=True,
                                       local_site=local_site,
                                       local_id=1002))
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                2)

        # Fetch the count for the LocalSite.
        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                2)

        # Fetch the count for the global site. This should still be in cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

    def test_unstar_review_request_with_discarded(self):
        """Testing Profile.unstar_review_request with discarded review request
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(local_site=None)

        review_request = self.create_review_request(publish=True)
        review_request.close(ReviewRequest.DISCARDED)
        profile.starred_review_requests.add(review_request)

        # 2 queries:
        #
        # 1. Fetch existing starred entries (to avoid duplicates)
        # 2. Insert new entry
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'where': (Q(profile=(profile.pk,)) &
                          Q(reviewrequest__in={review_request.pk})),
            },
            {
                'type': 'DELETE',
                'model': Profile.starred_review_requests.through,
                'where': Q(id__in=[review_request.pk]),
            },
        ]

        with assert_queries(queries):
            profile.unstar_review_request(review_request)

        self.assertFalse(profile.starred_review_requests.exists())
        self.assertEqual(site_profile.starred_public_request_count, 0)
        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): False,
            })

    def test_unstar_review_request_with_draft(self):
        """Testing Profile.unstar_review_request with draft review request"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(local_site=None)

        review_request = self.create_review_request()
        profile.starred_review_requests.add(review_request)

        # 2 queries:
        #
        # 1. Fetch existing starred entries (to avoid duplicates)
        # 2. Insert new entry
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'where': (Q(profile=(profile.pk,)) &
                          Q(reviewrequest__in={review_request.pk})),
            },
            {
                'type': 'DELETE',
                'model': Profile.starred_review_requests.through,
                'where': Q(id__in=[review_request.pk]),
            },
        ]

        with assert_queries(queries):
            profile.unstar_review_request(review_request)

        self.assertFalse(profile.starred_review_requests.exists())
        self.assertEqual(site_profile.starred_public_request_count, 0)
        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): False,
            })

    def test_unstar_review_request_with_pending(self):
        """Testing Profile.unstar_review_request with pending review request"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(local_site=None)

        review_request = self.create_review_request(publish=True)
        profile.starred_review_requests.add(review_request)
        site_profile.starred_public_request_count = 1
        site_profile.save(update_fields=('starred_public_request_count',))

        # 4 queries:
        #
        # 1. Fetch existing starred entries (to avoid duplicates)
        # 2. Remove the entry
        # 3. Update LocalSiteProfile.starred_public_request_count
        # 4. Re-fetch LocalSiteProfile.starred_public_request_count
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'where': (Q(profile=(profile.pk,)) &
                          Q(reviewrequest__in={review_request.pk})),
            },
            {
                'type': 'DELETE',
                'model': Profile.starred_review_requests.through,
                'where': Q(id__in=[review_request.pk]),
            },
            {
                'type': 'UPDATE',
                'model': LocalSiteProfile,
                'where': Q(pk=site_profile.pk),
            },
            {
                'model': LocalSiteProfile,
                'values_select': ('starred_public_request_count',),
                'where': Q(pk=site_profile.pk),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            profile.unstar_review_request(review_request)

        self.assertFalse(profile.starred_review_requests.exists())
        self.assertEqual(site_profile.starred_public_request_count, 0)
        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): False,
            })

    def test_unstar_review_request_with_submitted(self):
        """Testing Profile.unstar_review_request with submitted review request
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(local_site=None)

        review_request = self.create_review_request(publish=True)
        review_request.close(ReviewRequest.SUBMITTED)

        profile.starred_review_requests.add(review_request)
        site_profile.starred_public_request_count = 1
        site_profile.save(update_fields=('starred_public_request_count',))

        # 4 queries:
        #
        # 1. Fetch existing starred entries (to avoid duplicates)
        # 2. Remove the entry
        # 3. Update LocalSiteProfile.starred_public_request_count
        # 4. Re-fetch LocalSiteProfile.starred_public_request_count
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'where': (Q(profile=(profile.pk,)) &
                          Q(reviewrequest__in={review_request.pk})),
            },
            {
                'type': 'DELETE',
                'model': Profile.starred_review_requests.through,
                'where': Q(id__in=[review_request.pk]),
            },
            {
                'type': 'UPDATE',
                'model': LocalSiteProfile,
                'where': Q(pk=site_profile.pk),
            },
            {
                'model': LocalSiteProfile,
                'values_select': ('starred_public_request_count',),
                'where': Q(pk=site_profile.pk),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            profile.unstar_review_request(review_request)

        self.assertFalse(profile.starred_review_requests.exists())
        self.assertEqual(site_profile.starred_public_request_count, 0)
        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): False,
            })

    @add_fixtures(['test_site'])
    def test_unstar_review_request_invalidates_cache(self):
        """Testing Profile.unstar_review_request invalidates count cache"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)
        local_site.users.add(user)

        review_request1 = self.create_review_request()
        review_request2 = self.create_review_request()
        profile.star_review_request(review_request1)
        profile.star_review_request(review_request2)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        cross_site_queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                2)

        # Fetch the count for the global site.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        global_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 2)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 2)

        # Fetch the count for the LocalSite.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        local_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                0)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                0)

        # Unstar it. This should invalidate the global and cross-site caches.
        profile.unstar_review_request(review_request2)

        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): True,
                (ReviewRequest, 2): False,
            })

        # Fetch the count for cross-sites.
        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                1)

        # Fetch the count for the global site.
        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 1)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 1)

        # Fetch the count for the LocalSite. This should still be in cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                0)

    @add_fixtures(['test_site'])
    def test_unstar_review_request_with_local_site_invalidates_cache(self):
        """Testing Profile.unstar_review_request with LocalSite invalidates
        count cache
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)
        local_site.users.add(user)

        review_request1 = self.create_review_request(local_site=local_site,
                                                     local_id=1)
        review_request2 = self.create_review_request(local_site=local_site,
                                                     local_id=2)
        profile.star_review_request(review_request1)
        profile.star_review_request(review_request2)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # Fetch the count for cross-sites.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        cross_site_queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                2)

        # Fetch the count for the LocalSite.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        local_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                2)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                2)

        # Fetch the count for the global site.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review request count.
        global_site_queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(global_site_queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

        # Unstar it. This should invalidate the LocalSite and cross-site
        # caches.
        profile.unstar_review_request(review_request2)

        self.assertTrue(LocalSite.objects.has_local_sites())
        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): True,
                (ReviewRequest, 2): False,
            })

        # Fetch the count for cross-sites.
        with assert_queries(cross_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                1)

        # Fetch the count for the LocalSite.
        with assert_queries(local_site_queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                1)

        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                1)

        # Fetch the count for the global site. This should still be in cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

    def test_get_starred_review_groups_count(self):
        """Testing Profile.get_starred_review_groups_count"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertFalse(LocalSite.objects.has_local_sites())

        # This should start out as 0.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 0)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 0)

        profile.starred_groups.add(
            self.create_review_group(name='group1'),
            self.create_review_group(name='group2'))
        cache.clear()

        self.assertFalse(LocalSite.objects.has_local_sites())

        # This should now be 2.
        with assert_queries(queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 2)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 2)

    @add_fixtures(['test_site'])
    def test_get_starred_review_groups_count_with_local_site_in_db(self):
        """Testing Profile.get_starred_review_groups_count with LocalSites in
        database
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should start out as 0.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 0)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 0)

        profile.starred_groups.add(
            self.create_review_group(name='group1'),
            self.create_review_group(name='group2'),
            self.create_review_group(name='group3',
                                     with_local_site=True))
        cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should now be 2.
        with assert_queries(queries):
            self.assertEqual(profile.get_starred_review_groups_count(), 2)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_groups_count(), 2)

    @add_fixtures(['test_site'])
    def test_get_starred_review_groups_count_with_local_site(self):
        """Testing Profile.get_starred_review_groups_count with local_site="""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should start out as 0.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                0)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                0)

        # Star review groups and invalidate cache.
        profile.starred_groups.add(
            self.create_review_group(name='group1'),
            self.create_review_group(name='group2'),
            self.create_review_group(name='group3',
                                     with_local_site=True))
        cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should now be 1.
        with assert_queries(queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                1)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(local_site=local_site),
                1)

    @add_fixtures(['test_site'])
    def test_get_starred_review_groups_count_with_local_site_all(self):
        """Testing Profile.get_starred_review_groups_count with
        local_site=LocalSite.ALL
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This should start out as 0.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                0)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                0)

        # Star review groups and invalidate cache.
        profile.starred_groups.add(
            self.create_review_group(name='group1'),
            self.create_review_group(name='group2'),
            self.create_review_group(name='group3',
                                     with_local_site=True))
        cache.clear()

        # This should now be 1.
        with assert_queries(queries):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                3)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_groups_count(
                    local_site=LocalSite.ALL),
                3)

    def test_get_starred_review_requests_count(self):
        """Testing Profile.get_starred_review_requests_count"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertFalse(LocalSite.objects.has_local_sites())

        # This should start out as 0.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

        # Star review requests and invalidate cache.
        profile.starred_review_requests.add(
            self.create_review_request(),
            self.create_review_request())
        cache.clear()

        self.assertFalse(LocalSite.objects.has_local_sites())

        # This should now be 2.
        with assert_queries(queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 2)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 2)

    @add_fixtures(['test_site'])
    def test_get_starred_review_requests_count_with_local_site_in_db(self):
        """Testing Profile.get_starred_review_requests_count with LocalSites
        in database
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should start out as 0.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 0)

        # Star review requests and invalidate cache.
        profile.starred_review_requests.add(
            self.create_review_request(),
            self.create_review_request(),
            self.create_review_request(with_local_site=True))
        cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should now be 2.
        with assert_queries(queries):
            self.assertEqual(profile.get_starred_review_requests_count(), 2)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(profile.get_starred_review_requests_count(), 2)

    @add_fixtures(['test_site'])
    def test_get_starred_review_requests_count_with_local_site(self):
        """Testing Profile.get_starred_review_requests_count with local_site=
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should start out as 0.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                0)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                0)

        # Star review requests and invalidate cache.
        profile.starred_review_requests.add(
            self.create_review_request(),
            self.create_review_request(),
            self.create_review_request(with_local_site=True))
        cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should now be 1.
        with assert_queries(queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                1)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=local_site),
                1)

    @add_fixtures(['test_site'])
    def test_get_starred_review_requests_count_with_local_site_all(self):
        """Testing Profile.get_starred_review_requests_count with
        local_site=LocalSite.ALL
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This should start out as 0.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                0)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                0)

        # Star review requests and invalidate cache.
        profile.starred_review_requests.add(
            self.create_review_request(),
            self.create_review_request(),
            self.create_review_request(with_local_site=True))
        cache.clear()

        # This should now be 1.
        with assert_queries(queries):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                3)

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                profile.get_starred_review_requests_count(
                    local_site=LocalSite.ALL),
                3)

    def test_has_starred_review_groups(self):
        """Testing Profile.has_starred_review_groups"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertFalse(LocalSite.objects.has_local_sites())

        # This should start out as False.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(profile.has_starred_review_groups())

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(profile.has_starred_review_groups())

        # Star groups and invalidate cache.
        profile.starred_groups.add(
            self.create_review_group(name='group1'),
            self.create_review_group(name='group2'))
        cache.clear()

        self.assertFalse(LocalSite.objects.has_local_sites())

        # This should now be True.
        with assert_queries(queries):
            self.assertTrue(profile.has_starred_review_groups())

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertTrue(profile.has_starred_review_groups())

    @add_fixtures(['test_site'])
    def test_has_starred_review_groups_with_local_site_in_db(self):
        """Testing Profile.has_starred_review_groups with LocalSites in
        database
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should start out as False.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(profile.has_starred_review_groups())

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(profile.has_starred_review_groups())

        # Star groups and invalidate cache.
        profile.starred_groups.add(
            self.create_review_group(name='group1'),
            self.create_review_group(name='group2'),
            self.create_review_group(name='group3',
                                     with_local_site=True))
        cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should now be True.
        with assert_queries(queries):
            self.assertTrue(profile.has_starred_review_groups())

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertTrue(profile.has_starred_review_groups())

    @add_fixtures(['test_site'])
    def test_has_starred_review_groups_with_local_site(self):
        """Testing Profile.has_starred_review_groups with local_site="""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should start out as False.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(
                profile.has_starred_review_groups(local_site=local_site))

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(
                profile.has_starred_review_groups(local_site=local_site))

        # Star groups and invalidate cache.
        profile.starred_groups.add(
            self.create_review_group(name='group1'),
            self.create_review_group(name='group2'),
            self.create_review_group(name='group3',
                                     with_local_site=True))
        cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should now be True.
        with assert_queries(queries):
            self.assertTrue(
                profile.has_starred_review_groups(local_site=local_site))

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertTrue(
                profile.has_starred_review_groups(local_site=local_site))

    @add_fixtures(['test_site'])
    def test_has_starred_review_groups_with_local_site_all(self):
        """Testing Profile.has_starred_review_groups with
        local_site=LocalSite.ALL
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This should start out as False.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(
                profile.has_starred_review_groups(local_site=LocalSite.ALL))

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(
                profile.has_starred_review_groups(local_site=LocalSite.ALL))

        # Star groups and invalidate cache.
        profile.starred_groups.add(
            self.create_review_group(name='group1'),
            self.create_review_group(name='group2'),
            self.create_review_group(name='group2',
                                     with_local_site=True))
        cache.clear()

        # This should now be True.
        with assert_queries(queries):
            self.assertTrue(
                profile.has_starred_review_groups(local_site=LocalSite.ALL))

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertTrue(
                profile.has_starred_review_groups(local_site=LocalSite.ALL))

    def test_has_starred_review_requests(self):
        """Testing Profile.has_starred_review_requests"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertFalse(LocalSite.objects.has_local_sites())

        # This should start out as False.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(profile.has_starred_review_requests())

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(profile.has_starred_review_requests())

        # Star review requests and invalidate cache.
        profile.starred_review_requests.add(
            self.create_review_request(),
            self.create_review_request())
        cache.clear()

        self.assertFalse(LocalSite.objects.has_local_sites())

        # This should now be True.
        with assert_queries(queries):
            self.assertTrue(profile.has_starred_review_requests())

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertTrue(profile.has_starred_review_requests())

    @add_fixtures(['test_site'])
    def test_has_starred_review_requests_with_local_site_in_db(self):
        """Testing Profile.has_starred_review_requests with LocalSites in
        database
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should start out as False.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(profile.has_starred_review_requests())

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(profile.has_starred_review_requests())

        # Star review requests and invalidate cache.
        profile.starred_review_requests.add(
            self.create_review_request(),
            self.create_review_request(),
            self.create_review_request(with_local_site=True))
        cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should now be True.
        with assert_queries(queries):
            self.assertTrue(profile.has_starred_review_requests())

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertTrue(profile.has_starred_review_requests())

    @add_fixtures(['test_site'])
    def test_has_starred_review_requests_with_local_site(self):
        """Testing Profile.has_starred_review_requests with local_site="""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        local_site = self.get_local_site(self.local_site_name)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should start out as False.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=local_site),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(
                profile.has_starred_review_requests(local_site=local_site))

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(
                profile.has_starred_review_requests(local_site=local_site))

        # Star review requests and invalidate cache.
        profile.starred_review_requests.add(
            self.create_review_request(),
            self.create_review_request(),
            self.create_review_request(with_local_site=True))
        cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # This should now be True.
        with assert_queries(queries):
            self.assertTrue(
                profile.has_starred_review_requests(local_site=local_site))

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertTrue(
                profile.has_starred_review_requests(local_site=local_site))

    @add_fixtures(['test_site'])
    def test_has_starred_review_requests_with_local_site_all(self):
        """Testing Profile.has_starred_review_requests with
        local_site=LocalSite.ALL
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        # This should start out as False.
        #
        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(
                profile.has_starred_review_requests(local_site=LocalSite.ALL))

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(
                profile.has_starred_review_requests(local_site=LocalSite.ALL))

        # Star review requests and invalidate cache.
        profile.starred_review_requests.add(
            self.create_review_request(),
            self.create_review_request(),
            self.create_review_request(with_local_site=True))
        cache.clear()

        # This should now be True.
        with assert_queries(queries):
            self.assertTrue(
                profile.has_starred_review_requests(local_site=LocalSite.ALL))

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertTrue(
                profile.has_starred_review_requests(local_site=LocalSite.ALL))

    def test_is_review_request_starred(self):
        """Testing Profile.is_review_request_starred"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        review_request = self.create_review_request()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertFalse(LocalSite.objects.has_local_sites())

        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile)
            },
        ]

        with assert_queries(queries):
            self.assertFalse(profile.is_review_request_starred(review_request))

        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): False,
            })

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(profile.is_review_request_starred(review_request))

        # Star a review request and invalidate cache.
        profile.starred_review_requests.add(review_request)
        cache.clear()
        profile.star_cache.clear()

        self.assertFalse(LocalSite.objects.has_local_sites())

        # 2 queries:
        #
        # 1. Fetch user's starred review requests count.
        # 2. starred_review_requests lookup
        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile)
            },
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {
                    'a': Value(1),
                },
                'where': (Q(profile=profile.pk) &
                          Q(reviewrequest=review_request.pk)),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            self.assertTrue(profile.is_review_request_starred(review_request))

        # These will be in cache.
        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): True,
            })

        # Another call will utilize star_cache.
        with self.assertNumQueries(0):
            self.assertTrue(profile.is_review_request_starred(review_request))

        # If we flush the cache, a new call will still perform the
        # starred_review_requests lookup, but not the count.
        profile.star_cache.clear()

        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {
                    'a': Value(1),
                },
                'where': (Q(profile=profile.pk) &
                          Q(reviewrequest=review_request.pk)),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            self.assertTrue(profile.is_review_request_starred(review_request))

    @add_fixtures(['test_site'])
    def test_is_review_request_starred_with_local_site_in_db(self):
        """Testing Profile.is_review_request_starred with LocalSites in
        database
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        review_request = self.create_review_request()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # 1 query:
        #
        # 1. Fetch user's starred review requests count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(profile.is_review_request_starred(review_request))

        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): False,
            })

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(profile.is_review_request_starred(review_request))

        # Star a review request and invalidate cache.
        profile.starred_review_requests.add(review_request)
        cache.clear()
        profile.star_cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # 2 queries:
        #
        # 1. Fetch user's starred review requests count.
        # 2. starred_review_requests lookup
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {
                    'a': Value(1),
                },
                'where': (Q(profile=profile.pk) &
                          Q(reviewrequest=review_request.pk)),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            self.assertTrue(profile.is_review_request_starred(review_request))

        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, 1): True,
            })

        # Another call will utilize star_cache.
        with self.assertNumQueries(0):
            self.assertTrue(profile.is_review_request_starred(review_request))

        # If we flush the cache, a new call will still perform the
        # starred_review_requests lookup, but not the count.
        profile.star_cache.clear()

        queries = [
            {
                'model': Profile.starred_review_requests.through,
                'annotations': {
                    'a': Value(1),
                },
                'where': (Q(profile=profile.pk) &
                          Q(reviewrequest=review_request.pk)),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            self.assertTrue(profile.is_review_request_starred(review_request))

    def test_is_review_group_starred(self):
        """Testing Profile.is_review_group_starred"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        review_group = self.create_review_group()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertFalse(LocalSite.objects.has_local_sites())

        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(profile.is_review_group_starred(review_group))

        self.assertEqual(
            profile.star_cache,
            {
                (Group, 1): False,
            })

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(profile.is_review_group_starred(review_group))

        # Star a review group and invalidate cache.
        profile.starred_groups.add(review_group)
        cache.clear()
        profile.star_cache.clear()

        self.assertFalse(LocalSite.objects.has_local_sites())

        # 2 queries:
        #
        # 1. The cache fetch
        # 2. starred_groups lookup
        queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {'__count': Count('*')},
                'where': Q(profile=profile),
            },
            {
                'model': Profile.starred_groups.through,
                'annotations': {
                    'a': Value(1),
                },
                'where': (Q(profile=profile.pk) &
                          Q(group=review_group.pk)),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            self.assertTrue(profile.is_review_group_starred(review_group))

        self.assertEqual(
            profile.star_cache,
            {
                (Group, 1): True,
            })

        # Another call will utilize star_cache.
        with self.assertNumQueries(0):
            self.assertTrue(profile.is_review_group_starred(review_group))

        # If we flush the cache, a new call will still perform the
        # starred_groups lookup, but not the count.
        profile.star_cache.clear()

        queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {
                    'a': Value(1),
                },
                'where': (Q(profile=profile.pk) &
                          Q(group=review_group.pk)),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            self.assertTrue(profile.is_review_group_starred(review_group))

    @add_fixtures(['test_site'])
    def test_is_review_group_starred_with_local_site_in_db(self):
        """Testing Profile.is_review_group_starred with LocalSites in database
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        review_group = self.create_review_group()

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # 1 query:
        #
        # 1. Fetch user's starred review groups count.
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(profile.is_review_group_starred(review_group))

        self.assertEqual(
            profile.star_cache,
            {
                (Group, 1): False,
            })

        # A second call should hit the cache.
        with self.assertNumQueries(0):
            self.assertFalse(profile.is_review_group_starred(review_group))

        # Star a review group and invalidate cache.
        profile.starred_groups.add(review_group)
        cache.clear()
        profile.star_cache.clear()

        self.assertTrue(LocalSite.objects.has_local_sites())

        # 2 queries:
        #
        # 1. The cache fetch
        # 2. starred_groups lookup
        queries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'annotations': {'__count': Count('*')},
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'where': Q(starred_by__id=user.pk) & Q(local_site=None),
            },
            {
                'model': Profile.starred_groups.through,
                'annotations': {
                    'a': Value(1),
                },
                'where': (Q(profile=profile.pk) &
                          Q(group=review_group.pk)),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            self.assertTrue(profile.is_review_group_starred(review_group))

        self.assertEqual(
            profile.star_cache,
            {
                (Group, 1): True,
            })

        # Another call will utilize star_cache.
        with self.assertNumQueries(0):
            self.assertTrue(profile.is_review_group_starred(review_group))

        # If we flush the cache, a new call will still perform the
        # starred_groups lookup, but not the count.
        profile.star_cache.clear()

        queries = [
            {
                'model': Profile.starred_groups.through,
                'annotations': {
                    'a': Value(1),
                },
                'where': (Q(profile=profile.pk) &
                          Q(group=review_group.pk)),
                'limit': 1,
            },
        ]

        with assert_queries(queries):
            self.assertTrue(profile.is_review_group_starred(review_group))

    def test_prefetch_starred_review_groups(self) -> None:
        """Testing Profile.prefetch_starred_review_groups"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        review_group1 = self.create_review_group(name='group1')
        review_group2 = self.create_review_group(name='group2')
        review_group3 = self.create_review_group(name='group3')

        profile.starred_groups.add(review_group2, review_group3)

        equeries: ExpectedQueries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'values_select': ('pk',),
                'where': (
                    Q(starred_by__id=user.pk) &
                    Q(pk__in=[review_group1.pk, review_group2.pk])
                ),
            },
        ]

        with assert_queries(equeries):
            profile.prefetch_starred_review_groups([
                review_group1.pk,
                review_group2.pk,
            ])

        # A second query for the cached items should be a no-op.
        with self.assertNumQueries(0):
            profile.prefetch_starred_review_groups([
                review_group1.pk,
                review_group2.pk,
            ])

        # A pre-fetch with an uncached entry should query again.
        equeries = [
            {
                'join_types': {
                    'accounts_profile_starred_groups': 'INNER JOIN',
                },
                'model': Group,
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_groups',
                    'reviews_group',
                },
                'values_select': ('pk',),
                'where': (
                    Q(starred_by__id=user.pk) &
                    Q(pk__in=[review_group3.pk])
                ),
            },
        ]

        with assert_queries(equeries):
            profile.prefetch_starred_review_groups([
                review_group3.pk,
            ])

        self.assertEqual(
            profile.star_cache,
            {
                (Group, review_group1.pk): False,
                (Group, review_group2.pk): True,
                (Group, review_group3.pk): True,
            })

    def test_prefetch_starred_review_requests(self) -> None:
        """Testing Profile.prefetch_starred_review_requests"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()

        review_request1 = self.create_review_request()
        review_request2 = self.create_review_request()
        review_request3 = self.create_review_request()

        profile.starred_review_requests.add(review_request2, review_request3)

        equeries: ExpectedQueries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'values_select': ('pk',),
                'where': (
                    Q(starred_by__id=user.pk) &
                    Q(pk__in=[review_request1.pk, review_request2.pk])
                ),
            },
        ]

        with assert_queries(equeries):
            profile.prefetch_starred_review_requests([
                review_request1.pk,
                review_request2.pk,
            ])

        # A second query for the cached items should be a no-op.
        with self.assertNumQueries(0):
            profile.prefetch_starred_review_requests([
                review_request1.pk,
                review_request2.pk,
            ])

        # A pre-fetch with an uncached entry should query again.
        equeries = [
            {
                'join_types': {
                    'accounts_profile_starred_review_requests': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'accounts_profile_starred_review_requests',
                    'reviews_reviewrequest',
                },
                'values_select': ('pk',),
                'where': (
                    Q(starred_by__id=user.pk) &
                    Q(pk__in=[review_request3.pk])
                ),
            },
        ]

        with assert_queries(equeries):
            profile.prefetch_starred_review_requests([
                review_request3.pk,
            ])

        self.assertEqual(
            profile.star_cache,
            {
                (ReviewRequest, review_request1.pk): False,
                (ReviewRequest, review_request2.pk): True,
                (ReviewRequest, review_request3.pk): True,
            })



"""Model re-exports for reviewboard.diffviewer.*."""

from reviewboard.diffviewer.models.diffcommit import DiffCommit
from reviewboard.diffviewer.models.diffset import DiffSet
from reviewboard.diffviewer.models.diffset_history import DiffSetHistory
from reviewboard.diffviewer.models.filediff import FileDiff
from reviewboard.diffviewer.models.legacy_file_diff_data import \
    LegacyFileDiffData
from reviewboard.diffviewer.models.raw_file_diff_data import RawFileDiffData


__all__ = [
    'DiffCommit',
    'DiffSet',
    'DiffSetHistory',
    'FileDiff',
    'LegacyFileDiffData',
    'RawFileDiffData',
]

__autodoc_excludes__ = __all__



"""Hosting service for Forgejo.

Version Added:
    8.0
"""

from __future__ import annotations

from reviewboard.hostingsvcs.forgejo.service import Forgejo


__all__ = [
    'Forgejo',
]

__autodoc_excludes__ = __all__



"""Backends for Single Sign-On.

Version Added:
    5.0
"""

from djblets.registries.importer import lazy_import_registry


#: The SSO backends registry
sso_backends = lazy_import_registry(
    'reviewboard.accounts.sso.backends.registry',
    'SSOBackendRegistry')



"""Extension hooks for augmenting e-mail messages."""

from __future__ import annotations

from collections import defaultdict
from typing import Tuple, TYPE_CHECKING

from django.dispatch import Signal

from reviewboard.notifications.email.utils import RecipientList
from reviewboard.reviews.signals import (review_request_published,
                                         review_published, reply_published,
                                         review_request_closed)

if TYPE_CHECKING:
    from reviewboard.extensions.hooks import EmailHook


# A mapping of signals to EmailHooks.
_hooks = defaultdict(set)


def register_email_hook(
    signal: Signal,
    handler: EmailHook,
) -> None:
    """Register an e-mail hook.

    Args:
        signal (django.dispatch.Signal):
            The signal that will trigger the e-mail to be sent. This is one of
            :py:data:`~reviewboard.reviews.signals.review_request_published`,
            :py:data:`~reviewboard.reviews.signals.review_request_closed`,
            :py:data:`~reviewboard.reviews.signals.review_published`, or
            :py:data:`~reviewboard.reviews.signals.reply_published`.

        handler (reviewboard.extensions.hooks.EmailHook):
            The ``EmailHook`` that will be triggered when an e-mail of the
            chosen type is about to be sent.
    """
    assert signal in (review_request_published, review_request_closed,
                      review_published, reply_published), (
        'Invalid signal %r' % signal)

    _hooks[signal].add(handler)


def unregister_email_hook(
    signal: Signal,
    handler: EmailHook,
) -> None:
    """Unregister an e-mail hook.

    Args:
        signal (django.dispatch.Signal):
            The signal that will trigger the e-mail to be sent. This is one of
            :py:data:`~reviewboard.reviews.signals.review_request_published`,
            :py:data:`~reviewboard.reviews.signals.review_request_closed`,
            :py:data:`~reviewboard.reviews.signals.review_published`, or
            :py:data:`~reviewboard.reviews.signals.reply_published`.

        handler (reviewboard.extensions.hooks.EmailHook):
            The ``EmailHook`` that will be triggered when an e-mail of the
            chosen type is about to be sent.
    """
    assert signal in (review_request_published, review_request_closed,
                      review_published, reply_published), (
        'Invalid signal %r' % signal)

    _hooks[signal].discard(handler)


def filter_email_recipients_from_hooks(
    to_field: RecipientList,
    cc_field: RecipientList,
    signal: Signal,
    **kwargs,
) -> Tuple[RecipientList, RecipientList]:
    """Filter the e-mail recipients through configured e-mail hooks.

    Args:
        to_field (set):
            The original To field of the e-mail, as a set of
            :py:class:`Users <django.contrib.auth.models.User>` and
            :py:class:`Groups <reviewboard.reviews.models.Group>`.

        cc_field (set):
            The original CC field of the e-mail, as a set of
            :py:class:`Users <django.contrib.auth.models.User>` and
            :py:class:`Groups <reviewboard.reviews.models.Group>`.

        signal (django.dispatch.Signal):
            The signal that triggered the e-mail.

        **kwargs (dict):
            Extra keyword arguments to pass to the e-mail hook.

    Returns:
        tuple:
        A 2-tuple of the To field and the CC field, as sets of
        :py:class:`Users <django.contrib.auth.models.User>` and
        :py:class:`Groups <reviewboard.reviews.models.Group>`.
    """
    if signal in _hooks:
        for hook in _hooks[signal]:
            to_field = hook.get_to_field(to_field, **kwargs)
            cc_field = hook.get_cc_field(cc_field, **kwargs)

    return to_field, cc_field



"""Internal utilities to help with database installs and upgrades.

These functions should not be called outside of :command:`rb-site` or
:file:`manage.py`.

Version Added:
    5.0
"""

# NOTE: Imports for anything that touches Django/database functionality must
#       be done within the upgrade function. Importing at the module level
#       can cause problems with apps being loaded in the wrong order.

from __future__ import annotations

from typing import Dict, List, Set, TYPE_CHECKING, Tuple, Type

from django.db import DatabaseError
from typing_extensions import TypedDict

if TYPE_CHECKING:
    from django.db.models import Model
    from reviewboard.cmdline.utils.console import Console


class UpgradeStateError(Exception):
    """An error indicating a problem with computed upgrade state.

    Version Added:
        5.0.2
    """


class UpgradeState(TypedDict, total=False):
    """State required for performing an upgrade.

    These are computed during the pre-upgrade steps, and used in the
    post-upgrade steps to determine what changes need to be made to the
    database or configurations.

    Version Added:
        5.0.2
    """

    #: A set of all table names currently in the database.
    #:
    #: This is always available to all pre/post-upgrade steps.
    #:
    #: Type:
    #:     set
    tables: Set[str]

    #: Whether SCMTool IDs need to be migrated.
    #:
    #: This is set when upgrading from pre-5.0 to 5.0 or higher.
    #:
    #: Type:
    #:     bool
    needs_scmtool_id_migration: bool

    #: A mapping of modern SCMTool IDs to repositories using the tool.
    #:
    #: This can only be set if :py:attr:`needs_scmtool_id_migration` is set.
    #:
    #: Type:
    #:     dict
    scmtool_id_data: Dict[str, List[int]]

    #: A mapping of legacy Tool PKs to modern SCMTool IDs.
    #:
    #: This can only be set if :py:attr:`needs_scmtool_id_migration` is set.
    #:
    #: Type:
    #:     dict
    tool_pk_to_scmtool_id: Dict[int, str]

    #: A set of IntegrationConfig PKs with conditions to update for SCMTools.
    #:
    #: This can only be set if :py:attr:`needs_scmtool_id_migration` is set.
    #:
    #: Type:
    #:     set
    conditions_for_scmtool_migration: Set[int]


def _had_model(
    upgrade_state: UpgradeState,
    model_cls: Type[Model],
) -> bool:
    """Return whether a model was installed in the pre-upgrade database.

    Version Added:
        5.0.2

    Args:
        upgrade_state (dict):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.

        console (reviewboard.cmdline.utils.console.Console, unused):
            The console output wrapper.

    Returns:
        bool:
        ``True`` if the model was installed pre-upgrade. ``False`` if it was
        not.
    """
    table_names = upgrade_state.get('tables', set())

    # As of Jan 6, 2023, pyright fails to find attributes on Type[Model].
    # In this case, Model._meta. To avoid issues, ignore the type.
    return model_cls._meta.db_table in table_names  # type: ignore


def pre_upgrade_gather_db_state(
    upgrade_state: UpgradeState,
    console: Console,
) -> None:
    """Store a list of all table names in the database.

    These are used by subsequent steps to better calculate which pieces of
    the database need to be upgraded or queried.

    Version Added:
        5.0.2

    Args:
        upgrade_state (dict):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.

        console (reviewboard.cmdline.utils.console.Console, unused):
            The console output wrapper.
    """
    from django.db import connection

    with connection.cursor() as cursor:
        upgrade_state['tables'] = \
            set(connection.introspection.table_names(cursor))


def pre_upgrade_reset_oauth2_provider(
    upgrade_state: UpgradeState,
    console: Console,
) -> None:
    """Reset the OAuth2 migration/evolution state pre-upgrade.

    This will remove any migration information regarding the
    :py:mod:`oauth2_provider` Django app from the list of migrations and the
    Django Evolution project signature. This allows Django Evolution to take
    control of the upgrade process again.

    Version Added:
        5.0

    Args:
        upgrade_state (dict, unused):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.

            This is not used by this pre-upgrade step.

        console (reviewboard.cmdline.utils.console.Console, unused):
            The console output wrapper.

    Raises:
        django.db.DatabaseError:
            There was an error querying the database.
    """
    from django.db import connection
    from django_evolution.consts import UpgradeMethod
    from django_evolution.models import Evolution, Version
    from django_evolution.utils.migrations import unrecord_applied_migrations

    if not _had_model(upgrade_state, Version):
        # There's no table. This is a brand-new installation. We have nothing
        # to do at this stage.
        return

    try:
        version = Version.objects.current_version()
    except Version.DoesNotExist:
        # There's no recorded versions. This is a brand-new installation.
        # We have nothing to do at this stage.
        return
    except DatabaseError as e:
        console.error('Unexpected error trying to query the "%s" table: %s'
                      % (Version._meta.db_table, e))
        raise

    project_sig = version.signature
    app_sig = project_sig.get_app_sig('oauth2_provider')

    if app_sig is None or app_sig.upgrade_method != UpgradeMethod.MIGRATIONS:
        return

    # Patch the signature to reset this app back to Django Evolution, undoing
    # the MoveToMigrations introduced in Review Board 4.
    app_sig.upgrade_method = UpgradeMethod.EVOLUTIONS
    app_sig.applied_migrations = []
    version.save()

    # Remove any migrations. We'll later record the migrations we want to
    # simulate having applied in the post-upgrade step.
    unrecord_applied_migrations(connection=connection,
                                app_label='oauth2_provider')

    # Remove the "move_to_migrations" evolution.
    evolutions = Evolution.objects.filter(app_label='oauth2_provider',
                                          label='move_to_migrations')
    evolutions.delete()


def post_upgrade_reset_oauth2_provider(
    upgrade_state: UpgradeState,
    console: Console,
) -> None:
    """Mark oauth2_provider migrations as applied.

    Post-upgrade, this will mark all oauth2_provider migrations as applied,
    satisfying migration dependency checks and Django startup checks.

    These steps will be executed regardless of whether we changed any state
    in :py:func:`pre_upgrade_reset_oauth2_provider`, since we always need to
    mirror the current version's migrations regardless of whether we've
    altered the database signature.

    This list of migrations must be updated whenever we update oauth2_provider.

    Version Added:
        5.0

    Args:
        upgrade_state (dict, unused):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.

            This is not used by this post-upgrade step.

        console (reviewboard.cmdline.utils.console.Console, unused):
            The console output wrapper.
    """
    from django.db import connection
    from django_evolution.utils.migrations import (MigrationList,
                                                   record_applied_migrations,
                                                   unrecord_applied_migrations)

    unrecord_applied_migrations(connection=connection,
                                app_label='oauth2_provider')

    # This is current as of oauth2_provider 1.6.3.
    record_applied_migrations(
        connection=connection,
        migrations=MigrationList.from_names(
            app_label='oauth2_provider',
            migration_names=[
                '0001_initial',
                '0002_auto_20190406_1805',
                '0003_auto_20201211_1314',
                '0004_auto_20200902_2022',
                '0005_auto_20211222_2352',
            ]))


def pre_upgrade_store_scmtool_data(
    upgrade_state: UpgradeState,
    console: Console,
) -> None:
    """Store the data for adding scmtool_id to the Repository object.

    Version Added:
        5.0

    Args:
        upgrade_state (dict):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.

        console (reviewboard.cmdline.utils.console.Console):
            The console output wrapper.

    Raises:
        django.db.DatabaseError:
            There was an error querying the database.
    """
    from django_evolution.models import Evolution
    from reviewboard.scmtools.models import Repository

    if not _had_model(upgrade_state, Evolution):
        # The evolution tables don't yet exist. This is a brand-new
        # installation. We have nothing to do at this stage.
        return

    try:
        evolution = Evolution.objects.filter(
            app_label='scmtools', label='repository_scmtool_id')

        has_evolution = evolution.exists()
    except DatabaseError as e:
        console.error('Unexpected error trying to query for the '
                      '"scmtools.repository_scmtool_id" evolution: %s'
                      % e)
        raise

    # We'll need to store state for an upgrade if we haven't yet added the
    # Repository.scmtool_id field or if we have any repositories remaining
    # that haven't been updated yet.
    try:
        needs_upgrade = (
            not has_evolution or
            Repository.objects.filter(scmtool_id=None).exists()
        )
    except Exception as e:
        # Something went wrong trying to perform this query. The evolution
        # may be in the database but the table hasn't been upgraded. Assume
        # an upgrade is needed.
        console.error('Unexpected error trying to determine if an upgrade '
                      'is required. Proceeding with the upgrade. Contact '
                      'Beanbag Support and report the following error if '
                      'you encounter any problems: %s'
                      % e)
        needs_upgrade = True

    if needs_upgrade:
        # TODO: We do eventually want to delete the Tool model entirely. When
        # we do that, this will need to change to run some hand-written SQL
        # because we won't have the Python-side available, even when the
        # table is still present in the database.
        from django.db.models import Prefetch, Q
        from reviewboard.scmtools import scmtools_registry
        from reviewboard.scmtools.models import Tool

        # This will just be 2 queries in total, optimized only for the fields
        # we need, and leveraging the database as best as possible:
        repositories_queryset = Repository.objects.only('pk', 'tool_id')

        if has_evolution:
            repositories_queryset = repositories_queryset.filter(
                Q(scmtool_id=None) &
                Q(tool_id__isnull=False))

        tools = (
            Tool.objects
            .prefetch_related(Prefetch(
                'repositories',
                queryset=repositories_queryset))
            .order_by('pk')
        )

        upgrade_state['needs_scmtool_id_migration'] = True

        scmtool_id_data: Dict[str, List[int]] = {}
        missing_tools: List[Tuple[str, Exception]] = []

        for tool in tools:
            try:
                scmtool_id = tool.scmtool_id
            except Exception as e:
                missing_tools.append((tool.name, e))
                continue

            # This is required instead of values_list() due to the prefetch,
            # but will be optimized due to what we chose to prefetch.
            scmtool_id_data[scmtool_id] = [
                repository.pk
                for repository in tool.repositories.all()
            ]

        errors: List[List[str]] = []

        if missing_tools:
            errors.append([
                'The following tools were registered in your database but '
                'could not be loaded due the following errors:',
            ] + [
                '* %s: %s' % (tool_name, e)
                for tool_name, e in missing_tools
            ] + [
                'For now, these tools are being skipped. Review requests '
                'using associated repositories may crash. Please ensure the '
                'proper packages are installed correctly, and then '
                're-upgrade the site directory.',
            ])

        if scmtools_registry.conflicting_tools:
            errors.append([
                'The following SCMTools in your database have been modified '
                'or renamed, and may no longer work correctly:',
            ] + [
                '* Your %s (%s) conflicts with our %s (%s)'
                % (conflict_tool.name,
                   conflict_tool.class_name,
                   scmtool_cls.name,
                   scmtool_cls.class_name)
                for (scmtool_cls,
                     conflict_tool) in scmtools_registry.conflicting_tools
            ] + [
                'If you are using custom SCMTools, you will '
                'need to register yours via an extension and update any '
                'repositories.',
            ])

        if errors:
            for error in errors:
                console.warning('\n'.join(error))

            console.note(
                'Contact Beanbag Support (support@beanbaginc.com) if you '
                'need help.',
            )

        upgrade_state['scmtool_id_data'] = scmtool_id_data
    else:
        upgrade_state['needs_scmtool_id_migration'] = False


def post_upgrade_apply_scmtool_data(
    upgrade_state: UpgradeState,
    console: Console,
) -> None:
    """Apply the scmtool_id migration data.

    Version Added:
        5.0

    Args:
        upgrade_state (dict):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.

        console (reviewboard.cmdline.utils.console.Console, unused):
            The console output wrapper.

    Raises:
        UpgradeStateError:
            The upgrade state had missing or bad data.
    """
    if not upgrade_state.get('needs_scmtool_id_migration'):
        return

    from reviewboard.scmtools.models import Repository

    try:
        scmtool_id_data = upgrade_state['scmtool_id_data']
    except KeyError:
        raise UpgradeStateError(
            '`scmtool_id_data` is missing in the pre-upgrade state!')

    for scmtool_id, repository_ids in scmtool_id_data.items():
        repositories = Repository.objects.filter(pk__in=repository_ids)
        repositories.update(scmtool_id=scmtool_id)


def pre_upgrade_store_condition_tool_info(
    upgrade_state: UpgradeState,
    console: Console,
) -> None:
    """Store the data for converting RepositoryTypeChoice data.

    The :py:class:`reviewboard.scmtools.conditions.RepositoryTypeChoice`
    traditionally used the Tool pk as its value. This upgrade sequence will
    convert those to use the SCMTool ID instead.

    This must run after :py:func:`pre_upgrade_store_scmtool_data`.

    Version Added:
        5.0

    Args:
        upgrade_state (dict):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.

        console (reviewboard.cmdline.utils.console.Console, unused):
            The console output wrapper.
    """
    from reviewboard.integrations.models import IntegrationConfig

    if (not upgrade_state.get('needs_scmtool_id_migration') or
        not _had_model(upgrade_state, IntegrationConfig)):
        # This doesn't need SCMTool ID migration, so we can skip this step.
        return

    from reviewboard.scmtools.models import Tool

    # This was a Review Board 3.0+ installation.
    tool_pks: Set[int] = set()
    affected_configs: Set[int] = set()

    configs = IntegrationConfig.objects.only('pk', 'settings')

    for config in configs:
        # We need to directly work with the underlying settings dictionary.
        # If we go through config.get(), and the setting doesn't exist, it
        # will result in looking up the integration instance, which may not
        # be available at this point (due to being part of an extension) or
        # may not even be installed, and that will crash the upgrade
        # process.
        conditions = config.settings.get('conditions')

        if conditions:
            for condition in conditions['conditions']:
                if condition['choice'] == 'repository_type':
                    changed: bool = False

                    for tool_pk in condition['value']:
                        try:
                            tool_pks.add(int(tool_pk))
                            changed = True
                        except ValueError:
                            # This was data from an already-converted
                            # condition. We don't want to try querying with
                            # it.
                            pass

                    if changed:
                        affected_configs.add(config.pk)

    tools = Tool.objects.filter(pk__in=tool_pks).only('pk', 'class_name')

    upgrade_state['tool_pk_to_scmtool_id'] = {
        tool.pk: tool.scmtool_id
        for tool in tools
    }
    upgrade_state['conditions_for_scmtool_migration'] = affected_configs


def post_upgrade_apply_condition_tool_info(
    upgrade_state: UpgradeState,
    console: Console,
) -> None:
    """Convert RepositoryTypeChoice conditions to use SCMTool ID.

    The :py:class:`reviewboard.scmtools.conditions.RepositoryTypeChoice`
    traditionally used the Tool pk as its value. This upgrade sequence will
    convert those to use the SCMTool ID instead.

    Version Added:
        5.0

    Args:
        upgrade_state (dict):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.

        console (reviewboard.cmdline.utils.console.Console, unused):
            The console output wrapper.

    Raises:
        UpgradeStateError:
            The upgrade state had missing or bad data.
    """
    from reviewboard.integrations.models import IntegrationConfig

    if (not upgrade_state.get('needs_scmtool_id_migration') or
        not _had_model(upgrade_state, IntegrationConfig)):
        # This doesn't need SCMTool ID migration, so we can skip this step.
        return

    try:
        tool_pk_to_scmtool_id = upgrade_state['tool_pk_to_scmtool_id']
        config_pks = upgrade_state['conditions_for_scmtool_migration']
    except KeyError as e:
        raise UpgradeStateError(
            '`%s` is missing in the pre-upgrade state!'
            % e)

    configs = (
        IntegrationConfig.objects
        .filter(pk__in=config_pks)
        .only('pk', 'settings')
    )

    for config in configs:
        # See the note above about how we access the settings for
        # configurations.
        #
        # It should be safe to use .get(), since we've already determined
        # the correct integrations above, but as this is the upgrade
        # process we'll want to be extra careful.
        conditions = config.settings.get('conditions')

        if conditions:
            for condition in conditions['conditions']:
                if condition['choice'] == 'repository_type':
                    condition['value'] = [
                        tool_pk_to_scmtool_id.get(pk, pk)
                        for pk in condition['value']
                    ]

            config.set('conditions', conditions)
            config.save(update_fields=('settings',))


def run_pre_upgrade_tasks(
    upgrade_state: UpgradeState,
    console: Console,
) -> None:
    """Run any database pre-upgrade tasks.

    Version Added:
        5.0

    Args:
        upgrade_state (dict):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.
            Pre-upgrade steps can modify this to include any information
            needed.

        console (reviewboard.cmdline.utils.console.Console):
            The console output wrapper.

    Raises:
        django.db.DatabaseError:
            There was an error querying the database.
    """
    pre_upgrade_gather_db_state(upgrade_state, console)
    pre_upgrade_reset_oauth2_provider(upgrade_state, console)
    pre_upgrade_store_scmtool_data(upgrade_state, console)
    pre_upgrade_store_condition_tool_info(upgrade_state, console)


def run_post_upgrade_tasks(
    upgrade_state: UpgradeState,
    console: Console,
) -> None:
    """Run any database post-upgrade tasks.

    Version Added:
        5.0

    Args:
        upgrade_state (dict):
            Upgrade state that can be used by pre-upgrade/post-upgrade steps.

        console (reviewboard.cmdline.utils.console.Console):
            The console output wrapper.

    Raises:
        UpgradeStateError:
            The upgrade state had missing or bad data.

        django.db.DatabaseError:
            There was an error querying the database.
    """
    post_upgrade_reset_oauth2_provider(upgrade_state, console)
    post_upgrade_apply_scmtool_data(upgrade_state, console)
    post_upgrade_apply_condition_tool_info(upgrade_state, console)



"""Reduce max_length of FileAttachment.repo_revision from 512 to 64.

Version Added:
    2.0
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('FileAttachment', 'repo_revision', initial=None,
                max_length=64),
]



"""Base information for a review request and draft."""
from __future__ import annotations

import re
from collections.abc import Iterator
from typing import TYPE_CHECKING, cast

from django.db import models
from django.utils.translation import gettext_lazy as _
from djblets.db.fields import JSONField

from reviewboard.attachments.models import FileAttachmentHistory
from reviewboard.reviews.models.default_reviewer import DefaultReviewer

if TYPE_CHECKING:
    from reviewboard.attachments.models import FileAttachmentSequence
    from reviewboard.reviews.models.screenshot import Screenshot


class BaseReviewRequestDetails(models.Model):
    """Base information for a review request and draft.

    ReviewRequest and ReviewRequestDraft share a lot of fields and
    methods. This class provides those fields and methods for those
    classes.
    """

    MAX_SUMMARY_LENGTH = 300

    description = models.TextField(_("description"), blank=True)
    description_rich_text = models.BooleanField(
        _('description in rich text'),
        default=False)

    testing_done = models.TextField(_("testing done"), blank=True)
    testing_done_rich_text = models.BooleanField(
        _('testing done in rich text'),
        default=False)

    bugs_closed = models.CharField(_("bugs"), max_length=300, blank=True)
    branch = models.CharField(_("branch"), max_length=300, blank=True)
    commit_id = models.CharField(_('commit ID'), max_length=64, blank=True,
                                 null=True, db_index=True)

    extra_data = JSONField(null=True)

    # Deprecated and no longer used for new review requests as of 2.0.9.
    rich_text = models.BooleanField(_("rich text"), default=False)

    def get_review_request(self):
        raise NotImplementedError

    def get_bug_list(self):
        """Returns a list of bugs associated with this review request."""
        if self.bugs_closed == "":
            return []

        bugs = list(set(re.split(r"[, ]+", self.bugs_closed)))

        # First try a numeric sort, to show the best results for the majority
        # case of bug trackers with numeric IDs.  If that fails, sort
        # alphabetically.
        try:
            bugs.sort(key=int)
        except ValueError:
            bugs.sort()

        return bugs

    def get_screenshots(self) -> Iterator[Screenshot]:
        """Return a generator for all active screenshots.

        This includes all current screenshots, but not previous inactive ones.

        By accessing screenshots through this method, future review request
        lookups from the screenshots will be avoided.

        Yields:
            reviewboard.reviews.models.screenshot.Screenshot:
            A screenshot on the review request or draft.
        """
        if self.screenshots_count > 0:
            review_request = self.get_review_request()

            from reviewboard.reviews.models.screenshot import Screenshot

            for screenshot in self.screenshots.all():
                screenshot._review_request = review_request
                yield cast(Screenshot, screenshot)

    def get_inactive_screenshots(self) -> Iterator[Screenshot]:
        """Return a generator for all inactive screenshots.

        This only includes screenshots that were previously visible but
        have since been removed.

        By accessing screenshots through this method, future review request
        lookups from the screenshots will be avoided.

        Yields:
            reviewboard.reviews.models.screenshot.Screenshot:
            An inactive screenshot on the review request or draft.
        """
        if self.inactive_screenshots_count > 0:
            review_request = self.get_review_request()

            for screenshot in self.inactive_screenshots.all():
                screenshot._review_request = review_request
                yield screenshot

    def get_file_attachments(
        self,
        *,
        sort: bool = True,
    ) -> FileAttachmentSequence:
        """Return a list for all active file attachments.

        This includes all current file attachments, but not previous inactive
        ones.

        By accessing file attachments through this method, future review
        request lookups from the file attachments will be avoided.

        Version Changed:
            6.0:
            Added the ``sort`` argument.

        Args:
            sort (bool, optional):
                Whether to sort the file attachments by display position.

                This is the default.

                Version Added:
                    6.0

        Returns:
            list of reviewboard.attachments.models.FileAttachment:
            The active file attachments on the review request or draft.
        """
        def get_attachments(review_request):
            for file_attachment in self.file_attachments.all():
                file_attachment._review_request = review_request

                # Handle legacy entries which don't have an associated
                # FileAttachmentHistory entry.
                if (not file_attachment.is_from_diff and
                    file_attachment.attachment_history_id is None):
                    history = FileAttachmentHistory.objects.create(
                        display_position=FileAttachmentHistory
                            .compute_next_display_position(
                                review_request))

                    review_request.file_attachment_histories.add(history)

                    file_attachment.attachment_history = history
                    file_attachment.save(update_fields=['attachment_history'])

                yield file_attachment

        def get_display_position(attachment):
            if attachment.attachment_history_id is not None:
                return attachment.attachment_history.display_position
            else:
                return 0

        if self.file_attachments_count > 0:
            review_request = self.get_review_request()

            if sort:
                return sorted(get_attachments(review_request),
                              key=get_display_position)

            return list(get_attachments(review_request))
        else:
            return []

    def get_inactive_file_attachments(self):
        """Return a generator for all inactive file attachments.

        This only includes file attachments that were previously visible
        but have since been removed.

        By accessing file attachments through this method, future review
        request lookups from the file attachments will be avoided.

        Yields:
            reviewboard.attachments.models.FileAttachment:
            An inactive file attachment on the review request or draft.
        """
        if self.inactive_file_attachments_count > 0:
            review_request = self.get_review_request()

            for file_attachment in self.inactive_file_attachments.all():
                file_attachment._review_request = review_request
                yield file_attachment

    def add_default_reviewers(self):
        """Add default reviewers based on the diffset.

        This method goes through the DefaultReviewer objects in the database
        and adds any missing reviewers based on regular expression comparisons
        with the set of files in the diff.
        """
        if not self.repository:
            return

        diffset = self.get_latest_diffset()

        if not diffset:
            return

        match_default_reviewer_ids = []

        # This won't actually be queried until needed, since we're not
        # evaluating the queryset at this stage. That means we save a lookup
        # if the list of default reviewers is empty below.
        files = diffset.files.values_list('source_file', 'dest_file')

        default_reviewers = (
            DefaultReviewer.objects.for_repository(self.repository,
                                                   self.local_site)
            .only('pk', 'file_regex')
        )

        for default_reviewer in default_reviewers:
            try:
                regex = re.compile(default_reviewer.file_regex)
            except Exception:
                continue

            for source_file, dest_file in files:
                if regex.match(source_file or dest_file):
                    match_default_reviewer_ids.append(default_reviewer.pk)
                    break

        if not match_default_reviewer_ids:
            return

        # Get the list of users and groups across all matched default
        # reviewers. We'll fetch them directly from the ManyToMany tables,
        # to avoid extra queries. Django's m2m.add() methods will ensure no
        # duplicates are added, and that insertions aren't performed if not
        # needed.
        self.target_people.add(*(
            entry.user
            for entry in (
                DefaultReviewer.people.through.objects
                .filter(defaultreviewer_id__in=match_default_reviewer_ids,
                        user__is_active=True)
                .select_related('user')
            )
        ))

        self.target_groups.add(*(
            entry.group
            for entry in (
                DefaultReviewer.groups.through.objects
                .filter(defaultreviewer_id__in=match_default_reviewer_ids)
                .select_related('group')
            )
        ))

    def save(self, **kwargs):
        self.bugs_closed = self.bugs_closed.strip()
        self.summary = self._truncate(self.summary, self.MAX_SUMMARY_LENGTH)

        super(BaseReviewRequestDetails, self).save(**kwargs)

    def _truncate(self, string, num):
        if len(string) > num:
            string = string[0:num]
            i = string.rfind('.')

            if i != -1:
                string = string[0:i + 1]

        return string

    def __str__(self):
        if self.summary:
            return str(self.summary)
        else:
            return str(_('(no summary)'))

    class Meta:
        abstract = True
        app_label = 'reviews'



"""Managers for Web API related models.

Version Added:
    5.0.5
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Tuple, Union

from django.db.models import F, Q
from django.utils import timezone
from djblets.secrets.token_generators import token_generator_registry
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.symbols import UNSET, Unsettable
from djblets.webapi.managers import (
    WebAPITokenManager as DjbletsWebAPITokenManager)
from housekeeping import deprecate_non_keyword_only_args

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser
    from django.contrib.auth.models import AnonymousUser
    from django.db.models import QuerySet
    from typelets.json import JSONDict, JSONDictImmutable

    from reviewboard.webapi.models import WebAPIToken


class WebAPITokenManager(DjbletsWebAPITokenManager):
    """Manager for the reviewboard.webapi.managers.WebAPIToken model.

    Version Added:
        5.0.5
    """

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def get_or_create_client_token(
        self,
        *,
        user: Union[AbstractBaseUser, AnonymousUser],
        client_name: str,
        expires: Unsettable[datetime.datetime | None] = UNSET,
        min_validity_secs: int = 5 * 60,
        default_extra_data: Unsettable[JSONDictImmutable | None] = UNSET,
        default_policy: Unsettable[JSONDictImmutable | None] = UNSET,
        local_site: (LocalSite | None) = None,
        ignore_deprecated: bool = False,
    ) -> Tuple[WebAPIToken, bool]:
        """Return an API token for authenticating a client to Review Board.

        If the token does not already exist for the client, this will create
        one. If multiple client tokens already exist, this will return the
        one that has no expiration date or the furthest expiration date.
        If the eligible client token is expired or invalid, a new one will
        be created.

        By default, tokens with less than 5 minutes of validity left are
        excluded from results, which may cause another token to be used or
        a new token to be created. This helps avoid issues with tokens
        being returned that are immediately unusable. This validity window
        can be increased for longer-running operations.

        Version Changed:
            8.1:
            * By default, tokens with less than 5 minutes of validity left
              are excluded.

            * All arguments are now keyword-only. This will be required in
              Review Board 10.

            * Added the ``default_extra_data``, ``default_policy``,
              ``ignore_deprecated``, ``local_site``, and ``min_validity_secs``
              arguments.

        Args:
            user (django.contrib.auth.models.User):
                The user who owns the token.

            client_name (str):
                The name of the client that the token is for.

            expires (datetime.datetime, optional):
                The expiration datetime of new tokens.

                This defaults to the ``client_token_expiration`` value in the
                site configuration added to the creation datetime of the token.

                For example, if the ``client_token_expiration`` is set to 5,
                the token will expire in 5 days from when it is created.

            min_validity_secs (int, optional):
                The minimum number of seconds left in any existing token
                considered for a result.

                This defaults to 5 minutes.

                Version Added:
                    8.1

            default_extra_data (dict, optional):
                Extra data to set for new tokens.

                Version Added:
                    8.1

            default_policy (dict, optional):
                An explicit policy to set for new tokens.

                Version Added:
                    8.1

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site to associate with this token.

                Version Added:
                    8.1

            ignore_deprecated (bool, optional):
                Whether to ignore deprecated existing tokens.

                Version Added:
                    8.1

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (reviewboard.webapi.models.WebAPIToken):
                    The token for authenticating the client.

                1 (bool):
                    Whether a new token was created.

        Raises:
            djblets.webapi.errors.WebAPITokenGenerationError:
                The token was not able to be generated after the max
                number of collisions were hit.
        """
        valid_until = timezone.now()

        if min_validity_secs > 0:
            valid_until += datetime.timedelta(seconds=min_validity_secs)

        tokens: QuerySet[WebAPIToken] = (
            self.filter(
                Q(user=user) &
                Q(valid=True) &
                (Q(expires__isnull=True) |
                 Q(expires__gt=valid_until)) &
                LocalSite.objects.build_q(local_site=local_site)
            )
            .order_by(F('expires').desc(nulls_first=True))
        )

        for token in tokens:
            if (token.extra_data.get('client_name') == client_name and
                (not ignore_deprecated or not token.is_deprecated())):
                # This is an active token. Return it as-is.
                return token, False

        if expires is UNSET:
            siteconfig = SiteConfiguration.objects.get_current()
            expire_amount = siteconfig.get('client_token_expiration')

            if expire_amount and isinstance(expire_amount, int):
                expires = (timezone.now() +
                           datetime.timedelta(days=expire_amount))
            else:
                expires = None

        if default_policy is UNSET:
            policy = None
        else:
            policy = default_policy

        extra_data: JSONDict = {}

        if default_extra_data and default_extra_data is not UNSET:
            extra_data.update(default_extra_data)

        extra_data['client_name'] = client_name

        generator = token_generator_registry.get_default().token_generator_id
        token = self.generate_token(
            expires=expires,
            extra_data=extra_data,
            local_site=local_site,
            note=f'API token automatically created for {client_name}.',
            token_generator_id=generator,
            token_info={
                'token_type': 'rbp',
            },
            user=user,
            policy=policy,
        )

        return token, True



"""Template tags pertaining to the product as a whole."""

from django import template
from django.conf import settings
from djblets.util.templatetags.djblets_js import json_dumps

from reviewboard import (VERSION, get_manual_url, get_version_string,
                         is_release)


register = template.Library()

_product_info_str = None


@register.simple_tag
def js_product_info():
    """Return JSON-serialized information on the product.

    This will include the product name, version (human-readable and raw
    version information), release status, and the URL to the manual, in a
    form that can be directly embedded into a template's JavaScript section.

    Since the data won't change between calls without performing an upgrade
    and restart, result is cached in the process. Repeated calls will return
    the cached information.

    Returns:
        django.utils.safestring.SafeText:
        The JSON-serialized product information.
    """
    global _product_info_str

    # We're caching the results, since this isn't going to ever change
    # between requests while the process is still running.
    if _product_info_str is None:
        _product_info_str = json_dumps({
            'isRelease': is_release(),
            'manualURL': get_manual_url(),
            'name': settings.PRODUCT_NAME,
            'version': get_version_string(),
            'versionInfo': VERSION[:-1],
        })

    return _product_info_str



"""Forwarding imports and legacy utilities for authentication backends.

This module provides legacy functionality for registering authentication
backends, along with forwarding imports for:

.. autosummary::
   :nosignatures:

   ~reviewboard.accounts.backends.ad.ActiveDirectoryBackend
   ~reviewboard.accounts.backends.base.BaseAuthBackend
   ~reviewboard.accounts.backends.http_digest.HTTPDigestBackend
   ~reviewboard.accounts.backends.ldap.LDAPBackend
   ~reviewboard.accounts.backends.nis.NISBackend
   ~reviewboard.accounts.backends.registry.AuthBackendRegistry
   ~reviewboard.accounts.backends.registry.auth_backends
   ~reviewboard.accounts.backends.registry.get_enabled_auth_backends
   ~reviewboard.accounts.backends.standard.StandardAuthBackend
   ~reviewboard.accounts.backends.x509.X509Backend

Version Changed:
    3.0:
    The contents of this module were split into sub-modules.
"""

from reviewboard.accounts.backends.ad import ActiveDirectoryBackend
from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.backends.http_digest import HTTPDigestBackend
from reviewboard.accounts.backends.ldap import LDAPBackend
from reviewboard.accounts.backends.nis import NISBackend
from reviewboard.accounts.backends.registry import (AuthBackendRegistry,
                                                    auth_backends,
                                                    get_enabled_auth_backends)
from reviewboard.accounts.backends.standard import StandardAuthBackend
from reviewboard.accounts.backends.x509 import X509Backend


# Legacy references.
AuthBackend = BaseAuthBackend
INVALID_USERNAME_CHAR_REGEX = AuthBackend.INVALID_USERNAME_CHAR_REGEX


__all__ = (
    'ActiveDirectoryBackend',
    'AuthBackend',
    'AuthBackendRegistry',
    'HTTPDigestBackend',
    'INVALID_USERNAME_CHAR_REGEX',
    'LDAPBackend',
    'NISBackend',
    'StandardAuthBackend',
    'X509Backend',
    'auth_backends',
    'get_enabled_auth_backends',
)

__autodoc_excludes__ = (
    'ActiveDirectoryBackend',
    'AuthBackendRegistry',
    'HTTPDigestBackend',
    'LDAPBackend',
    'NISBackend',
    'StandardAuthBackend',
    'X509Backend',
    'auth_backends',
    'get_enabled_auth_backends',
)



"""API resource for managing hosting service accounts."""

from __future__ import annotations

from typing import Optional, Sequence, TYPE_CHECKING

from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_response_errors,
                                       webapi_request_fields)
from djblets.webapi.errors import (INVALID_FORM_DATA, NOT_LOGGED_IN,
                                   PERMISSION_DENIED)
from djblets.webapi.fields import IntFieldType, StringFieldType

from reviewboard.hostingsvcs.base import hosting_service_registry
from reviewboard.hostingsvcs.base.forms import BaseHostingServiceAuthForm
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.decorators import (webapi_check_login_required,
                                           webapi_check_local_site)
from reviewboard.webapi.errors import (BAD_HOST_KEY,
                                       HOSTINGSVC_AUTH_ERROR,
                                       REPO_AUTHENTICATION_ERROR,
                                       SERVER_CONFIG_ERROR,
                                       UNVERIFIED_HOST_CERT,
                                       UNVERIFIED_HOST_KEY)
from reviewboard.webapi.resources import resources

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from django.http import HttpRequest
    from djblets.webapi.responses import WebAPIResponseLinks
    from djblets.webapi.resources.base import WebAPIResourceHandlerResult


class HostingServiceAccountResource(WebAPIResource):
    """Provides information and allows linking of hosting service accounts.

    The list of accounts tied to hosting services can be retrieved, and new
    accounts can be linked through an HTTP POST.
    """

    item_resource_added_in = '1.6.7'
    list_resource_added_in = '2.5'

    name = 'hosting_service_account'
    model = HostingServiceAccount
    fields = {
        'id': {
            'type': IntFieldType,
            'description': 'The numeric ID of the hosting service account.',
        },
        'username': {
            'type': StringFieldType,
            'description': 'The username of the account.',
        },
        'service': {
            'type': StringFieldType,
            'description': 'The ID of the service this account is on.',
        },
    }
    uri_object_key = 'account_id'

    allowed_methods = ('GET', 'POST')

    item_child_resources = [
        resources.remote_repository,
    ]

    @webapi_check_login_required
    def get_queryset(
        self,
        request: HttpRequest,
        local_site_name: Optional[str] = None,
        is_list: bool = False,
        *args,
        **kwargs,
    ) -> QuerySet[HostingServiceAccount]:
        """Return a queryset for the resource.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            local_site_name (str, optional):
                The name of the current Local Site, if present.

            is_list (bool, optional):
                Whether to return a list queryset.

            *args (tuple):
                Positional arguments parsed from the URL.

            **kwargs (dict):
                Keyword arguments parsed from the URL.
        """
        local_site = self._get_local_site(local_site_name)

        queryset = self.model.objects.accessible(visible_only=True,
                                                 local_site=local_site)

        if is_list:
            if 'username' in request.GET:
                queryset = queryset.filter(username=request.GET['username'])

            if 'service' in request.GET:
                queryset = queryset.filter(service_name=request.GET['service'])

        return queryset

    def has_access_permissions(
        self,
        request: HttpRequest,
        account: HostingServiceAccount,
        *args,
        **kwargs,
    ) -> bool:
        """Return whether the account is accessible by a user.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The hosting service account to check.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            bool:
            Whether the hosting service account can be accessed by the user.
        """
        return account.is_accessible_by(request.user)

    def has_modify_permissions(
        self,
        request: HttpRequest,
        account: HostingServiceAccount,
        *args,
        **kwargs,
    ) -> bool:
        """Return whether the account is mutable by a user.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The hosting service account to check.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            bool:
            Whether the hosting service account can be modified by the user.
        """
        return account.is_mutable_by(request.user)

    def has_delete_permissions(
        self,
        request: HttpRequest,
        account: HostingServiceAccount,
        *args,
        **kwargs,
    ) -> bool:
        """Return whether the account can be deleted by a user.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The hosting service account to check.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            bool:
            Whether the hosting service account can be deleted by the user.
        """
        return account.is_mutable_by(request.user)

    def get_links(
        self,
        resources: Sequence[WebAPIResource] = [],
        obj: Optional[HostingServiceAccount] = None,
        *args,
        **kwargs,
    ) -> WebAPIResponseLinks:
        """Return links for the resource.

        Args:
            resources (list of reviewboard.webapi.base.WebAPIResource):
                A list of resources to include links to.

            obj (reviewboard.hostingsvcs.models.HostingServiceAccount,
                 optional):
                The current hosting service account, if accessing an item
                resource.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            djblets.webapi.responses.WebAPIResponseLinks:
            The links to include in the payload.
        """
        links = super().get_links(resources, obj=obj, *args, **kwargs)

        if obj:
            service = obj.service

            if not service.supports_list_remote_repositories:
                del links['remote_repositories']

        return links

    @webapi_request_fields(optional={
        'username': {
            'type': StringFieldType,
            'description': 'Filter accounts by username.',
            'added_in': '2.5',
        },
        'service': {
            'type': StringFieldType,
            'description': 'Filter accounts by the hosting service ID.',
            'added_in': '2.5',
        },
    })
    @augment_method_from(WebAPIResource)
    def get_list(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> WebAPIResourceHandlerResult:
        """Retrieves the list of accounts on the server.

        This will only list visible accounts. Any account that the
        administrator has hidden will be excluded from the list.
        """
        ...

    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs) -> WebAPIResourceHandlerResult:
        """Retrieves information on a particular account.

        This will only return very basic information on the account.
        Authentication information is not provided.
        """
        ...

    def serialize_service_field(
        self,
        obj: HostingServiceAccount,
        **kwargs,
    ) -> str:
        """Serialize the ``service`` field.

        Args:
            obj (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The hosting service account.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            str:
            The serialized content for the service field.
        """
        return obj.service_name

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(BAD_HOST_KEY, INVALID_FORM_DATA, NOT_LOGGED_IN,
                            PERMISSION_DENIED, REPO_AUTHENTICATION_ERROR,
                            SERVER_CONFIG_ERROR, UNVERIFIED_HOST_CERT,
                            UNVERIFIED_HOST_KEY)
    @webapi_request_fields(
        required={
            'username': {
                'type': StringFieldType,
                'description': 'The username on the account.',
            },
            'service_id': {
                'type': StringFieldType,
                'description': 'The registered ID of the service for the '
                               'account.',
            },
        },
        optional={
            'hosting_url': {
                'type': StringFieldType,
                'description': 'The hosting URL on the account, if the '
                               'hosting service is self-hosted.',
                'added_in': '1.7.8',
            },
            'password': {
                'type': StringFieldType,
                'description': 'The password on the account, if the hosting '
                               'service needs it.',
            },
        },
        allow_unknown=True,
    )
    def create(
        self,
        request: HttpRequest,
        username: str,
        service_id: str,
        *,
        password: Optional[str] = None,
        hosting_url: Optional[str] = None,
        local_site_name: Optional[str] = None,
        extra_fields: dict[str, str],
        **kwargs,
    ) -> WebAPIResourceHandlerResult:
        """Creates a hosting service account.

        The ``service_id`` is a registered HostingService ID. This must be
        known beforehand, and can be looked up in the Review Board
        administration UI.

        Depending on the hosting service, other parameters may be required.
        These can include API keys or two-factor auth tokens, and is dependent
        on each type of service. If a service requires additional fields,
        making requests to this API method will return an error indicating what
        fields are missing.
        """
        local_site = self._get_local_site(local_site_name)

        if not HostingServiceAccount.objects.can_create(request.user,
                                                        local_site):
            return self.get_no_access_error(request)

        # Validate the service.
        service = hosting_service_registry.get_hosting_service(service_id)

        if not service:
            return INVALID_FORM_DATA, {
                'fields': {
                    'service': ['This is not a valid service name'],
                },
            }

        if service.self_hosted and not hosting_url:
            return INVALID_FORM_DATA, {
                'fields': {
                    'hosting_url': ['This field is required'],
                },
            }

        if service.needs_authorization:
            form_cls = service.auth_form or BaseHostingServiceAuthForm
            form = form_cls(
                {
                    'hosting_account_password': password,
                    'hosting_account_username': username,
                    'hosting_url': hosting_url,
                    **extra_fields,
                },
                hosting_service_cls=service,
                local_site=local_site)

            if not form.is_valid():
                invalid_fields: dict[str, list[str]] = {}

                for field_name, errors in form.errors.items():
                    invalid_fields[field_name] = [
                        ', '.join(error.messages)
                        for error in errors.as_data()
                    ]

                return INVALID_FORM_DATA, {
                    'fields': invalid_fields,
                }

            try:
                account = form.save()
            except Exception as e:
                return HOSTINGSVC_AUTH_ERROR, {
                    'reason': str(e),
                }
        else:
            account = HostingServiceAccount(service_name=service_id,
                                            username=username,
                                            hosting_url=hosting_url,
                                            local_site=local_site)

            account.save()

        return 201, {
            self.item_result_key: account,
        }


hosting_service_account_resource = HostingServiceAccountResource()



"""Resource that links to watched items resources for a user."""

from __future__ import annotations

from typing import TYPE_CHECKING

from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.decorators import (
    webapi_check_local_site,
    webapi_check_login_required,
)
from reviewboard.webapi.resources import resources

if TYPE_CHECKING:
    from django.http import HttpRequest

    from djblets.webapi.resources.base import WebAPIResourceHandlerResult


class WatchedResource(WebAPIResource):
    """Resource that links to watched items resources for a user.

    This is more of a linking resource rather than a data resource, much like
    the root resource is. The sole purpose of this resource is for easy
    navigation to the more specific Watched Items resources.
    """

    name = 'watched'
    singleton = True

    list_child_resources = [
        resources.watched_review_group,
        resources.watched_review_request,
    ]

    @webapi_check_login_required
    @webapi_check_local_site
    def get(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> WebAPIResourceHandlerResult:
        """Retrieves the list of Watched Items resources.

        Unlike most resources, the result of this resource is just a list of
        links, rather than any kind of data. It exists in order to index the
        more specific Watched Review Groups and Watched Review Requests
        resources.
        """
        return 200, {
            'links': self.get_links(self.list_child_resources,
                                    request=request,
                                    *args, **kwargs),
        }


watched_resource = WatchedResource()



"""Unit tests for the FogBugz hosting service."""

from __future__ import annotations

from reviewboard.hostingsvcs.fogbugz import FogBugz
from reviewboard.hostingsvcs.testing import HostingServiceTestCase


class FogBugzTests(HostingServiceTestCase[FogBugz]):
    """Unit tests for the FogBugz hosting service."""

    service_name = 'fogbugz'
    fixtures = ['test_scmtools']

    def test_service_support(self) -> None:
        """Testing FogBugz service support capabilities"""
        self.assertTrue(self.service_class.supports_bug_trackers)
        self.assertFalse(self.service_class.supports_repositories)

    def test_get_bug_tracker_field(self) -> None:
        """Testing FogBugz.get_bug_tracker_field"""
        self.assertFalse(
            self.service_class.get_bug_tracker_requires_username())
        self.assertEqual(
            self.service_class.get_bug_tracker_field(None, {
                'fogbugz_account_domain': 'mydomain',
            }),
            'https://mydomain.fogbugz.com/f/cases/%s')



"""A hook for adding avatar services."""

from __future__ import annotations

from djblets.extensions.hooks import BaseRegistryHook, ExtensionHookPoint

from reviewboard.avatars import avatar_services


class AvatarServiceHook(BaseRegistryHook, metaclass=ExtensionHookPoint):
    """"A hook for adding avatar services.

    This hook will register services with the avatar services registry and
    unregister them when the hook is shut down.
    """

    registry = avatar_services

    def initialize(self, service):
        """Initialize the avatar service hook with the given service.

        Args:
            service (type):
                The avatar service class to register.

                This must be a subclass of
                :py:class:`djblets.avatars.services.base.AvatarService`.
        """
        super(AvatarServiceHook, self).initialize(service)



"""Testing utilities for building expected queries for repositories.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Dict, Set, TYPE_CHECKING, Union

from django.contrib.auth.models import User
from django.db.models import Q

from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries
    from django.contrib.auth.models import AnonymousUser

    from reviewboard.site.models import AnyOrAllLocalSites


def get_repositories_accessible_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: AnyOrAllLocalSites = None,
    visible_only: bool = True,
    distinct: bool = True,
) -> ExpectedQueries:
    """Return expected queries for Repository.objects.accessible().

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        visible_only (bool, optional):
            Whether the queries are for visible repositories only.

        distinct (bool, optional):
            Whether to expect a distinct query.

    Returns:
        list of dict:
        The list of expected queries.
    """
    # This is intentionally made to be verbose. Rather than constructing Q
    # objects more dynamically (like the accessible() implementation does),
    # we want to have each case build up the full query, so each combination
    # is explicitly defined.
    repositories_tables: Set[str] = {'scmtools_repository'}
    join_types: Dict[str, str] = {}

    if user.is_superuser:
        if visible_only:
            if local_site is LocalSite.ALL:
                repositories_where = Q(visible=True)
            else:
                repositories_where = (Q(visible=True) &
                                      Q(local_site=local_site))
        else:
            if local_site is LocalSite.ALL:
                repositories_where = Q()
            else:
                repositories_where = Q(local_site=local_site)
    elif user.is_authenticated:
        repositories_tables.update({
            'reviews_group',
            'reviews_group_users',
            'scmtools_repository_review_groups',
            'scmtools_repository_users',
        })
        join_types.update({
            'reviews_group': 'LEFT OUTER JOIN',
            'reviews_group_users': 'LEFT OUTER JOIN',
            'scmtools_repository_review_groups': 'LEFT OUTER JOIN',
            'scmtools_repository_users': 'LEFT OUTER JOIN',
        })

        if visible_only:
            if local_site is LocalSite.ALL:
                repositories_where = ((Q(public=True) &
                                       Q(visible=True)) |
                                      Q(users__pk=user.pk) |
                                      Q(review_groups__users=user.pk))
            else:
                repositories_where = (((Q(public=True) &
                                        Q(visible=True)) |
                                       Q(users__pk=user.pk) |
                                       Q(review_groups__users=user.pk)) &
                                      Q(local_site=local_site))
        else:
            if local_site is LocalSite.ALL:
                repositories_where = (Q(public=True) |
                                      (Q(users__pk=user.pk) |
                                       Q(review_groups__users=user.pk)))
            else:
                repositories_where = ((Q(public=True) |
                                       (Q(users__pk=user.pk) |
                                        Q(review_groups__users=user.pk))) &
                                      Q(local_site=local_site))
    else:
        assert user.is_anonymous

        if visible_only:
            if local_site is LocalSite.ALL:
                repositories_where = (Q(public=True) &
                                      Q(visible=True))
            else:
                repositories_where = (Q(public=True) &
                                      Q(visible=True) &
                                      Q(local_site=local_site))
        else:
            if local_site is LocalSite.ALL:
                repositories_where = Q(public=True)
            else:
                repositories_where = (Q(public=True) &
                                      Q(local_site=local_site))

    return [
        {
            '__note__': 'Fetch a list of accessible repositories',
            'distinct': distinct,
            'join_types': join_types,
            'model': Repository,
            'num_joins': len(repositories_tables) - 1,
            'tables': repositories_tables,
            'where': repositories_where,
        },
    ]


def get_repositories_accessible_ids_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: AnyOrAllLocalSites = None,
    visible_only: bool = True,
) -> ExpectedQueries:
    """Return expected queries for Repository.objects.accessible_ids().

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        visible_only (bool, optional):
            Whether the queries are for visible repositories only.

    Returns:
        list of dict:
        The list of expected queries.
    """
    equeries = get_repositories_accessible_equeries(
        user=user,
        local_site=local_site,
        visible_only=visible_only,
        distinct=False)
    equeries[-1].update({
        '__note__': 'Fetch a list of accessible repository IDs',
        'values_select': ('pk',),
    })

    return equeries



"""Integrations support for Review Board.

This module provides the Review Board functionality needed to create
integrations for third-party services. It builds upon Djblets's integrations
foundation, offering some additional utilities for more easily creating
manageable integrations.
"""



"""Rename Repository.password to Repository.encrypted_password.

Version Added:
    2.0.9
"""

from django_evolution.mutations import RenameField


MUTATIONS = [
    RenameField('Repository', 'password', 'encrypted_password',
                db_column='password'),
]



"""Admin definitions for change descriptions."""

from __future__ import annotations

from django.contrib import admin
from django.template.defaultfilters import truncatechars
from django.utils.translation import gettext_lazy as _, gettext

from reviewboard.admin import ModelAdmin, admin_site
from reviewboard.changedescs.models import ChangeDescription


class ChangeDescriptionAdmin(ModelAdmin):
    """Admin definitions for the ChangeDescription model."""

    list_display = ('truncated_text', 'public', 'timestamp')
    list_filter = ('timestamp', 'public')
    readonly_fields = ('fields_changed',)

    @admin.display(description=_('Change Description Text'))
    def truncated_text(
        self,
        obj: ChangeDescription,
    ) -> str:
        """Return the text of the object, truncated to 60 characters.

        Args:
            obj (reviewboard.changedescs.models.ChangeDescription):
                The change description.

        Returns:
            str:
            The text to show for the object.
        """
        text = obj.text

        if text:
            return truncatechars(obj.text, 60)
        else:
            return gettext('[empty description text]')


admin_site.register(ChangeDescription, ChangeDescriptionAdmin)
